#include <iostream.h>
#include <fstream.h>
#include <strstream.h>
#include <assert.h>
#include <vector.h>
#include <strings.h>

#include "general.h"
#include "Timer.h"
#include "Timer.icc"
#include "script.h"
#include "script.icc"

#define PIONEER_II

script::script( string& fileName,
		const char* refFileName = NULL,
		float timeFactor = 1.0,
		bool skipIfRobotStands = false,
		int skipTime = 0,
		int numberOfLaserScanners = 1,
		int scriptType = SCRIPT_TYPE_NOT_SET ) :
  laser(numberOfLaserScanners * 180),
#ifndef PIONEER_II
  sonar(24),
#else
  sonar(8),
#endif  
  _scriptType( scriptType),
  _stdin(NULL),
  _timeFactor(timeFactor),
  _skipIfRobotStands( skipIfRobotStands)
{
  if ( fileName == (string) "stdin") {
    _stdin = stdin;
    //    cerr << "# Get data from stdin.\n";
  }
#ifdef USE_ZLIB
  else if (( _zFile = gzopen( fileName.c_str(), "r")) == NULL) {
    cerr << "Cannot open " << fileName << "\n";
    //exit(1);
    return;
  }
#else
  else if (( _file = fopen( fileName.c_str(), "r")) == NULL) {
    cerr << "Cannot open " << fileName << "\n";
    //exit(1);
    return; 
  }
#endif
//   else
//     cerr << "# Get data from " << fileName << ".\n";

  bool timeRead = false;
  char line[maxStringLength];

  if ( _scriptType == SCRIPT_TYPE_NOT_SET) 
    while ( getNextLine( line) && (_scriptType == SCRIPT_TYPE_NOT_SET)) {
      if (readingLine(line, origScriptOpenMark)){
	_scriptType = OLD_SCRIPT_TYPE;
      }
      else if (readingLine(line, newScriptOpenMark)){
	_scriptType = NEW_SCRIPT_TYPE;
      }
    }

  if ( _scriptType == OLD_SCRIPT_TYPE) {
    positionMark = origScriptPosMark;
    timeMark = origScriptTimeMark;
    laserMark = origScriptLaserMark;
    sonarMark = origScriptSonarMark;
    imageMark = origScriptImageMark;
    timeOffset = origScriptTimeOffset;
    //    cerr << "# Old script" << endl;
  }
  else if ( _scriptType == NEW_SCRIPT_TYPE){
    positionMark = newScriptPosMark;
    timeMark = newScriptTimeMark;
    laserMark = newScriptLaserMark;
    sonarMark = newScriptSonarMark;
    imageMark = newScriptImageMark;
    timeOffset = newScriptTimeOffset;
    //    cerr << "# New script" << endl;
  }
  else {
    cerr << "# Error: Cannot determine script type.\n";
    exit(0);
  }

  if ( skipTime >= 0) {

    // First look for the start time of the script.

    while ( getNextLine( line) && !timeRead){

      if (readingLine(line, timeMark)){
	
	_startTime = getTime(line, timeMark, timeOffset);
	
	// Now skip the next skipTime seconds.
	if ( skipTime > 0) {
	  
	  Timer timeToBeReached = _startTime;
	  timeToBeReached.secs() += skipTime;
	  
	  cout << "# Found script start time: " << _startTime << endl;
	  
	  while ( getNextLine( line) && !timeRead){
	    if (readingLine(line, timeMark)){
	      _currentTime = getTime(line, timeMark, timeOffset);
	      if ( _currentTime >= timeToBeReached)
		timeRead = true;
	    }
	  }
	}
	else {
	  timeRead = true;
	  _currentTime = _startTime;
	}	
	//	cout << "# Start time: " << _currentTime << endl;
      }
    }
    
    assert( timeRead);
  }
  
  // Initialize the start time of the script and the real time of day.
  _realTime.reset();

  _elapsedTime = timeDiff( _currentTime, _startTime);
  _timeSkipped = _elapsedTime.secs() + _elapsedTime.usecs() / 1000000;

  laser.setOld();
  sonar.setOld();
  scriptPosition.x = 0.0;
  scriptPosition.y = 0.0;
  scriptPosition.rot = 0.0;
  scriptPosition.isNew = false;

  if ( refFileName )
    _refFile = fopen( refFileName, "r" );
  else
    _refFile = NULL;
  initRefPosition();
}

// initializes reference position
void 
script::initRefPosition() {
  refPosition.x = 0.0;
  refPosition.y = 0.0;
  refPosition.rot = 0.0;
}

//------------------------------------------------------------
// Gets the next sensor value from the script and sleeps until
// the time mark in the script is reached.
//------------------------------------------------------------
bool
script::getNextItem( bool considerTime = true)
{
  char line[maxStringLength];
  Timer newScriptTime = _currentTime;
  laser.setOld();
  sonar.setOld();
  scriptPosition.isNew = false;

  bool newSensor;

  // If "_skipIfRobotStands" is set, we search in the script for the next changed
  // robot position. Until then we don't want to wait.
  bool positionHasChanged = false;
  bool positionChangeSatisfied = ! _skipIfRobotStands;
  bool timeChangeSatisfied = false;
  float timeToSleep = 0.0;
  position nextPos;
  //float timeStamp;
  do {
    if ( getNextLine( line)){

      if (readingLine(line, sonarMark) ){
	sonar.getReadings(line);
      }
      else if (readingLine(line, laserMark)){
	laser.getReadings(line);
	// read in reference position if refFile exists
	if ( _refFile ) {
	  fscanf( _refFile, "%lf %lf %lf", &refPosition.x,               
		  &refPosition.y, &refPosition.rot );
    }
      }
      
      // Check whether the position has changed.
      else if (readingLine(line, positionMark)){

	nextPos = getPosition(line, positionMark);
	scriptPosition.isNew = true;
	if ( nextPos.x != scriptPosition.x ||
	     nextPos.y != scriptPosition.y ||
	     nextPos.rot != scriptPosition.rot) {
	  scriptPosition = nextPos;
	  positionHasChanged = true;
	}
	else {
	}
      }
      // Check whether a bump has occured.
      else if (readingLine(line, scriptBumpMark)){

	dumpBump( line, scriptBumpMark, _elapsedTime.floatTime());
      }
      else if ( readingLine(line, timeMark)){

	newScriptTime = getTime(line, timeMark, timeOffset);
	
	// Just forget about the time spent if the robot didn't move.
	if ( ! positionChangeSatisfied) {
	  _timeSkipped += newScriptTime - _currentTime;
	}
	
	_currentTime = newScriptTime;
	
	// Check whether the data is not too old and determine how long
	// the process should sleep.
	float realTimeExpired   = _realTime.expiredTime();
	float scriptTimeExpired = _currentTime - _startTime - _timeSkipped;
	scriptTimeExpired /= _timeFactor;
	timeToSleep = scriptTimeExpired - realTimeExpired;
      }
    }
    else
      return false;
    
    newSensor = sonar.isNew() || laser.isNew() || scriptPosition.isNew;
    timeChangeSatisfied     = ! considerTime || (timeToSleep > 0.0);
    positionChangeSatisfied = ! _skipIfRobotStands || positionHasChanged;
  }
  while ( ! (newSensor && positionChangeSatisfied && timeChangeSatisfied));
  
  _elapsedTime = timeDiff( _currentTime, _startTime);
  

  // Just read a new sensor value. Let's sleep until it is up to date.
#ifdef USE_ZLIB
  if ( ! gzeof( _zFile)) {
#else
    if ( ! feof( stdin)) {
#endif
      if ( timeToSleep > 0.0) { 
	if ( considerTime)
	  usleep( (uint) (timeToSleep * 1000000));
      }
      else if ( considerTime && timeToSleep < 0)
	cerr << "NEG" << timeToSleep << endl;
      _currentTime = newScriptTime;
    
    return true;
  }
  else
    return false;
}
